#! /usr/bin/env python

#################################################################################
# MIT License

# Copyright (c) 2023 Olaf Conrad

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#################################################################################
# Purpose
#################################################################################

'''
The PySAGA.helper module provides a collection of general helper functions.
'''


#################################################################################
#
# Globals
#________________________________________________________________________________

from PySAGA import saga_api


#################################################################################
#
# Prints...
#________________________________________________________________________________

#################################################################################
#________________________________________________________________________________
def Print_Table_Fields(Table):
    '''
    Print the attribute fields of a saga_api.CSG_Table object.

    Arguments
    ---------
    Table : the table object of type `saga_api`.`CSG_Table`

    Returns
    ---------
    boolean : True on success
    '''
    try:
        Table.Get_Field_Count()
    except:
        return False
    print()
    print('=====================')
    print('{:s} [{:d} fields, {:d} records]'.format(Table.Get_Name(), Table.Get_Field_Count(), Table.Get_Count()))
    print('_____________________')
    for Field in range(0, Table.Get_Field_Count()):
        print('{:2d}. {:s} [{:s}]'.format(Field + 1, Table.Get_Field_Name(Field), saga_api.SG_Data_Type_Get_Name(Table.Get_Field_Type(Field)).c_str()))
    print('=====================')
    return True


#################################################################################
#________________________________________________________________________________
def Print_Table(Table, Separator='\t', Fields=[], NoData='---', Maximum=0):
    '''
    Print the contents of a saga_api.CSG_Table object.

    Arguments
    ---------
    Table : the table object of type `saga_api`.`CSG_Table`
    Separator : `string` used to separate fields
    Fields : list with field indices (0 based) to be printed.
    Prints all fields if empty or None (default)
    NoData : prints this string for no-data values, or the
    original (no-data) value if `None`
    Maximum : maximum number of records to become printed.
    Prints all records if zero (default)

    Returns
    ---------
    boolean : True on success
    '''
    try:
        Table.Get_Field_Count()
    except:
        return False
    print()
    print('=====================')
    print('{:s} [{:d} fields, {:d} records]'.format(Table.Get_Name(), Table.Get_Field_Count(), Table.Get_Count()))
    print('_____________________')
    if Maximum < 1 or Maximum > Table.Get_Count():
        nRecords = Table.Get_Count()
    else:
        nRecords = Maximum
    if Fields and len(Fields) > 0:
        _Separator = ''
        for Field in Fields:
            if Field >= 0 and Field < Table.Get_Field_Count():
                print(_Separator + Table.Get_Field_Name(Field), end=''); _Separator = Separator
        print('\n_____________________')
        for i in range(0, nRecords):
            Record = Table.Get_Record(i); _Separator = ''
            for Field in Fields:
                if Field >= 0 and Field < Table.Get_Field_Count():
                    if NoData != None and Record.is_NoData(Field):
                        print(_Separator + NoData, end='')
                    else:
                        print(_Separator + Record.asString(Field), end='')
                    _Separator = Separator
            print()
    else:
        for Field in range(0, Table.Get_Field_Count()):
            if Field > 0:
                print(Separator, end='')
            print(Table.Get_Field_Name(Field), end='')
        print('\n_____________________')
        for i in range(0, nRecords):
            Record = Table.Get_Record(i)
            for Field in range(0, Table.Get_Field_Count()):
                if Field > 0:
                    print(Separator, end='')
                if NoData != None and Record.is_NoData(Field):
                    print(NoData, end='')
                else:
                    print(Record.asString(Field), end='')
            print()
    print('=====================')
    return True



#################################################################################
#
# Defining the Areas of Interest...
#________________________________________________________________________________

#################################################################################
#________________________________________________________________________________
def Get_AOI_From_Extent(Xmin, Xmax, Ymin, Ymax, EPSG=4326):
    '''
    Create area of interest from extent coordinates and EPSG code for CRS.
    '''

    AOI = saga_api.CSG_Shapes(saga_api.SHAPE_TYPE_Polygon)
    AOI.Get_Projection().Create(EPSG)
    Shape = AOI.Add_Shape()
    Shape.Add_Point(Xmin, Ymin)
    Shape.Add_Point(Xmin, Ymax)
    Shape.Add_Point(Xmax, Ymax)
    Shape.Add_Point(Xmax, Ymin)
    return AOI


#################################################################################
#________________________________________________________________________________
def Get_AOI_From_Features(File):
    '''
    Create area of interest from file. Expects to represent vector data with CRS
    information set correctly.
    '''

    AOI = saga_api.SG_Create_Shapes(File)
    if not AOI:
        saga_api.SG_UI_Msg_Add_Error('failed to load AOI from file \n\t\'{:s}\''.format(File))
        return None

    if not AOI.Get_Projection().is_Okay():
        del(AOI)
        saga_api.SG_UI_Msg_Add_Error('coordinate reference system of AOI is not defined \n\t\'{:s}\''.format(File))
        return None

    return AOI


#################################################################################
#________________________________________________________________________________
def Get_AOI_From_Raster(File):
    '''
    Create area of interest from file. Expects to represent raster data with CRS
    information set correctly.
    '''

    Grid = saga_api.SG_Create_Grid(File)
    if not Grid:
        saga_api.SG_UI_Msg_Add_Error('failed to load AOI from file \n\t\'{:s}\''.format(File))
        return None

    if not Grid.Get_Projection().is_Okay():
        del(Grid)
        saga_api.SG_UI_Msg_Add_Error('coordinate reference system of AOI is not defined \n\t\'{:s}\''.format(File))
        return None

    AOI = saga_api.CSG_Shapes(saga_api.SHAPE_TYPE_Polygon)
    AOI.Get_Projection().Create(Grid.Get_Projection())
    Shape = AOI.Add_Shape()
    Shape.Add_Point(Grid.Get_XMin(), Grid.Get_YMin())
    Shape.Add_Point(Grid.Get_XMin(), Grid.Get_YMax())
    Shape.Add_Point(Grid.Get_XMax(), Grid.Get_YMax())
    Shape.Add_Point(Grid.Get_XMax(), Grid.Get_YMin())
    del(Grid)
    return AOI


#################################################################################
#
# Virtual Raster Tiles...
#________________________________________________________________________________

#################################################################################
#________________________________________________________________________________
def Get_VRT(Directory, Extension = 'tif', VRT_Name = 'tiles', VRT_Folder = None):
    '''
    This function creates or updates the virtual raster (VRT) file named 'File'
    using all raster tiles with specified *Extension* found in *Directory*.
    If *VRT_Folder* is None resulting VRT file will be stored in the same directory
    as the tiles.
    Returns the full file path to the VRT file on success or None otherwise.
    '''

    import glob; Files = glob.glob('{:s}/*.{:s}'.format(Directory, Extension))
    if len(Files) < 1:
        saga_api.SG_UI_Console_Print_StdErr('directory \'{:s}\' does not contain any file with extension \'{:s}\'!'.format(Directory, Extension))
        return None

    Tiles = ''
    for File in Files:
        Tiles += '\"{:s}\" '.format(File)

    Tool = saga_api.SG_Get_Tool_Library_Manager().Get_Tool('io_gdal', '12')
    if not Tool:
        saga_api.SG_UI_Console_Print_StdErr('Failed to request tool: Create Virtual Raster (VRT)')
        return None

    if not VRT_Folder:
        VRT_Folder = Directory

    Tool.Reset()
    Tool.Set_Parameter('FILES'   , Tiles)
    Tool.Set_Parameter('VRT_NAME', '{:s}/{:s}.vrt'.format(VRT_Folder, VRT_Name))

    saga_api.SG_UI_ProgressAndMsg_Lock(True)
    if not Tool.Execute():
        saga_api.SG_UI_ProgressAndMsg_Lock(False)
        saga_api.SG_UI_Console_Print_StdErr('failed to execute tool: ' + Tool.Get_Name().c_str())
        return None

    saga_api.SG_UI_ProgressAndMsg_Lock(False)
    return Tool.Get_Parameter("VRT_NAME").asString()


#################################################################################
#
# SAGA / Python Type Conversations...
#________________________________________________________________________________

#________________________________________________________________________________
def PyList_from_Grid_Collection(Grids):
    '''
    Create a Python list that provides access to the single grids of the
    input grid collection.

    Arguments
    ---------
    Grids : expects a `grid collection` of type `saga_api`.`CSG_Grids`

    Returns
    ---------
    list : a list with `saga_api`.`CSG_Grid` objects
    '''

    List = []
    for i in range(0, Grids.Get_Grid_Count()):
        List.append(Grids.Get_Grid(i))
    return List

#________________________________________________________________________________
def PyList_to_Grid_Collection(List, Name='Grid Collection'):
    '''
    Create a grid collection from single grids provided by a Python list.

    Arguments
    ---------
    List : a Python list with objects of type `saga_api`.`CSG_Grid`

    Returns
    ---------
    `saga_api`.`CSG_Grids` : a grid collection of type `saga_api`.`CSG_Grids`
    '''

    Grids = saga_api.CSG_Grids()
    Grids.Fmt_Name(Name)
    for i in range(0, len(List)):
        Grids.Add_Grid(i, List[i].asGrid())
    return Grids


#################################################################################
#
# Tools and Tool Execution...
#________________________________________________________________________________

#________________________________________________________________________________
def Create_Toolboxes(Path='', SingleFile=False, UseToolName=True, Clean=True, LoadDefaults=False):
    '''
    PySAGA Tool Interface Generator
    ----------
    Create the Python interface for SAGA tools.

    Arguments
    ----------
    - Path [`string`] : either the single target file name or the target folder. If empty function tries to install targets in PySAGA directory (this file's directory).
    - SingleFile [`boolean`] : if `True` a single Python module will be created for all function calls, else one module for each tool library will be created
    - UseToolName [`boolean`]: if `True` function names will be based on the tool names, else these will named from `library` + `tool id`
    - Clean [`boolean`]: if `True` and not a single file is targeted all files from the target folder will be deleted before the new ones are generated
    - LoadDefaults [`boolean`]: if `True` libraries and tool chains will be loaded from default directories (if not done yet). Otherwise tool interfaces will only generated for the loaded libraries. Default libraries will also be loaded if no library has been loaded so far.
    Returns
    ----------
    `boolean` : `True` on success, `False` on failure.
    '''

    if not Path:
        import os; Path = os.path.dirname(__file__) + '/tools'
        if SingleFile:
            Path = Path + '.py'

    if LoadDefaults or saga_api.SG_Get_Tool_Library_Manager().Get_Count() < 1:
        saga_api.SG_Get_Tool_Library_Manager().Add_Default_Libraries()

    return saga_api.SG_Get_Tool_Library_Manager().Create_Python_ToolBox(saga_api.CSG_String(Path), Clean, UseToolName, SingleFile)


#################################################################################
#
# class Tool_Wrapper
#
#________________________________________________________________________________
class Tool_Wrapper:
    '''
    The Tool_Wrapper class is a smart wrapper for SAGA tools facilitating tool execution.
    '''

    Tool = None; Input = []; Output = []; Options = []

    #____________________________________
    def __init__(self, Library, Tool, Expected):
        self.Tool = saga_api.SG_Get_Tool_Library_Manager().Create_Tool(Library, Tool)
        if not self.Tool:
            saga_api.SG_UI_Msg_Add_Error('failed to request tool: {:s}'.format(Expected))
        else:
            self.Tool.Set_Manager(None)

    #____________________________________
    def __del__(self):
        self.Destroy()

    #____________________________________
    def Destroy(self):
        if self.is_Okay():
            saga_api.SG_Get_Tool_Library_Manager().Delete_Tool(self.Tool)
        self.Tool = None; self.Input.clear(); self.Output.clear(); self.Options.clear()

    #____________________________________
    def is_Okay(self):
        return self.Tool != None

    #____________________________________
    def Set_Input(self, ID, Object):
        if self.is_Okay() and Object != None:
            self.Input.append([ID, Object])

    #____________________________________
    def Set_Output(self, ID, Object):
        if self.is_Okay() and Object != None:
            self.Output.append([ID, Object])

    #____________________________________
    def Set_Option(self, ID, Value):
        if self.is_Okay() and Value != None:
            self.Options.append([ID, Value])

    #____________________________________
    def Execute(self, Verbose=2):
        if Verbose >= 2 and saga_api.SG_UI_Msg_is_Locked():
            Verbose = 1

        #________________________________
        def _Initialize():
            if Verbose <= 1:
                saga_api.SG_UI_ProgressAndMsg_Lock(True)
                if Verbose == 1:
                    saga_api.SG_UI_Console_Print_StdOut('[{:s}] execution... '.format(self.Tool.Get_Name().c_str()), '')

            for Option in self.Options: # 1st pass -> dis-/enabling data object parameters
                self.Tool.Set_Parameter(Option[0], Option[1])

            for Data in self.Input:
                Parameter = self.Tool.Get_Parameter(Data[0])
                if Parameter and Parameter.is_Input():
                    if Parameter.is_DataObject():
                        Parameter.Set_Value(Data[1])
                    elif Parameter.is_DataObject_List():
                        for Item in Data[1]:
                            Parameter.asList().Add_Item(Item)
                        Parameter.has_Changed()

            for Option in self.Options: # 2nd pass -> set options that are dependent of input data (-> table fields)
                self.Tool.Set_Parameter(Option[0], Option[1])

            for Data in self.Output:
                Parameter = self.Tool.Get_Parameter(Data[0])
                if Parameter and Parameter.is_Output() and Parameter.is_DataObject():
                    Parameter.Set_Value(Data[1])

            return self.Tool.On_Before_Execution()

        #________________________________
        def _Finalize(Success):
            self.Tool.On_After_Execution()

            for Data in self.Output:
                Parameter = self.Tool.Get_Parameter(Data[0])
                if Parameter and Parameter.is_Output():
                    if Parameter.is_DataObject():
                        if Parameter.asDataObject() and not Data[1]: # delete output created by tool but not requested by user
                            saga_api.SG_Data_Object_Delete(Parameter.asDataObject())
                    elif Parameter.is_DataObject_List():
                        if Data[1] != None:
                            Data[1].clear() # should be empty Python list to become filled with the results now
                        for i in range(0, Parameter.asList().Get_Item_Count()):
                            Item = Parameter.asList().Get_Item(i)
                            if Data[1] != None and Success:
                                Data[1].append(Item)
                            else:
                                saga_api.SG_Data_Object_Delete(Item)

            self.Tool.Reset(); self.Input.clear(); self.Output.clear()

            if Verbose <= 1:
                saga_api.SG_UI_ProgressAndMsg_Lock(False)
                if Verbose == 1:
                    if Success:
                        saga_api.SG_UI_Console_Print_StdOut('succeeded')
                    else:
                        saga_api.SG_UI_Console_Print_StdOut('failed')

        #________________________________
        if self.is_Okay():
            if _Initialize() and self.Tool.Execute():
                _Finalize(True)
                return True
            _Finalize(False)
        return False

#________________________________________________________________________________
#
# class Tool_Wrapper
#
#################################################################################


#################################################################################
#
#________________________________________________________________________________



#! /usr/bin/env python

'''
Python Interface to SAGA Tools Library
----------
- Category : Shapes
- Name     : Transects
- ID       : shapes_transect

Description
----------
A SAGA tool to create transects through polygon files.
'''

from PySAGA.helper import Tool_Wrapper

def Transect_through_polygon_shapefile(
    TRANSECT=None,
    THEME=None,
    TRANSECT_RESULT=None,
    THEME_FIELD=None, Verbose=2):
    '''
    Transect through polygon shapefile
    ----------
    [shapes_transect.0]\n
    Transect for lines and polygon shapefiles\n
    The goal of this tool is to create a transect along a line through a polygon map.\n
    Eg\n
    |____ST1_____!_ST2_!__ST1__!_______ST#_____|\n
    (Soil type 1 etc...)\n
    This is done by creating a table with the ID of each line, the distance\n
    to the starting point and the different transects:\n
    |  line_id  |  start  |  end  |  code/field  |\n
    |    0      |    0    |  124  |     ST1      |\n
    |    0      |   124   |  300  |     ST2      |\n
    |    0      |   300   | 1223  |     ST1      |\n
    |    0      |  1223   | 2504  |     ST3      |\n
    |    1      |    0    |  200  |     ST4      |\n
    |   ...     |   ...   |  ...  |     ...      |</pre>\n
    The tool requires an input shape with all the line transects [Transect_Line]\n
    and a polygon theme [Theme]. You also have to select which field you want to have in\n
    the resulting table [Transect_Result]. This can be an ID of the polygon theme if you\n
    want to link the tables later on, or any other field [Theme_Field].\n
    Arguments
    ----------
    - TRANSECT [`input shapes`] : Line Transect(s)
    - THEME [`input shapes`] : Theme
    - TRANSECT_RESULT [`output table`] : Result table
    - THEME_FIELD [`table field`] : Theme Field

    - Verbose [`integer number`] : Verbosity level, 0=silent, 1=tool name and success notification, 2=complete tool output.\n
    Returns
    ----------
    `boolean` : `True` on success, `False` on failure.
    '''
    Tool = Tool_Wrapper('shapes_transect', '0', 'Transect through polygon shapefile')
    if Tool.is_Okay():
        Tool.Set_Input ('TRANSECT', TRANSECT)
        Tool.Set_Input ('THEME', THEME)
        Tool.Set_Output('TRANSECT_RESULT', TRANSECT_RESULT)
        Tool.Set_Option('THEME_FIELD', THEME_FIELD)
        return Tool.Execute(Verbose)
    return False

def run_tool_shapes_transect_0(
    TRANSECT=None,
    THEME=None,
    TRANSECT_RESULT=None,
    THEME_FIELD=None, Verbose=2):
    '''
    Transect through polygon shapefile
    ----------
    [shapes_transect.0]\n
    Transect for lines and polygon shapefiles\n
    The goal of this tool is to create a transect along a line through a polygon map.\n
    Eg\n
    |____ST1_____!_ST2_!__ST1__!_______ST#_____|\n
    (Soil type 1 etc...)\n
    This is done by creating a table with the ID of each line, the distance\n
    to the starting point and the different transects:\n
    |  line_id  |  start  |  end  |  code/field  |\n
    |    0      |    0    |  124  |     ST1      |\n
    |    0      |   124   |  300  |     ST2      |\n
    |    0      |   300   | 1223  |     ST1      |\n
    |    0      |  1223   | 2504  |     ST3      |\n
    |    1      |    0    |  200  |     ST4      |\n
    |   ...     |   ...   |  ...  |     ...      |</pre>\n
    The tool requires an input shape with all the line transects [Transect_Line]\n
    and a polygon theme [Theme]. You also have to select which field you want to have in\n
    the resulting table [Transect_Result]. This can be an ID of the polygon theme if you\n
    want to link the tables later on, or any other field [Theme_Field].\n
    Arguments
    ----------
    - TRANSECT [`input shapes`] : Line Transect(s)
    - THEME [`input shapes`] : Theme
    - TRANSECT_RESULT [`output table`] : Result table
    - THEME_FIELD [`table field`] : Theme Field

    - Verbose [`integer number`] : Verbosity level, 0=silent, 1=tool name and success notification, 2=complete tool output.\n
    Returns
    ----------
    `boolean` : `True` on success, `False` on failure.
    '''
    Tool = Tool_Wrapper('shapes_transect', '0', 'Transect through polygon shapefile')
    if Tool.is_Okay():
        Tool.Set_Input ('TRANSECT', TRANSECT)
        Tool.Set_Input ('THEME', THEME)
        Tool.Set_Output('TRANSECT_RESULT', TRANSECT_RESULT)
        Tool.Set_Option('THEME_FIELD', THEME_FIELD)
        return Tool.Execute(Verbose)
    return False




#! /usr/bin/env python

'''
Python Interface to SAGA Tools Library
----------
- Category : Simulation
- Name     : Erosion
- ID       : sim_erosion

Description
----------
Modelling erosion processes.
'''

from PySAGA.helper import Tool_Wrapper

def MMFSAGA_Soil_Erosion_Model(
    DTM=None,
    S=None,
    PI=None,
    CC=None,
    PH=None,
    ETEO=None,
    GC=None,
    D=None,
    NV=None,
    MS=None,
    BD=None,
    EHD=None,
    LP=None,
    PER_C=None,
    PER_Z=None,
    PER_S=None,
    ST=None,
    CHANNEL=None,
    RFR=None,
    TAB_METEO=None,
    IF=None,
    Q=None,
    SL=None,
    Rf=None,
    KE=None,
    Rc=None,
    TCC=None,
    TCZ=None,
    TCS=None,
    Gc=None,
    Gz=None,
    Gs=None,
    SLC=None,
    SLZ=None,
    SLS=None,
    TCONDC=None,
    TCONDZ=None,
    TCONDS=None,
    W_UP=None,
    OUT_PATH=None,
    INTERFLOW=None,
    T=None,
    TIMESPAN=None,
    R=None,
    I=None,
    Rn=None,
    KE_I_METHOD=None,
    FLOWD_VA=None,
    CHANNELTRANSPORT=None, Verbose=2):
    '''
    MMF-SAGA Soil Erosion Model
    ----------
    [sim_erosion.0]\n
    Soil erosion modelling with a modified MMF (Morgan-Morgan-Finney) model (Morgan & Duzant 2008).[This tool is called MMF-SAGA because some things have been implemented differently compared to the original publication. The most important are:[[\n
    (-) the extension of the model to support spatially distributed modelling\n
    (-) the introduction of a "Channel Network" layer\n
    (-) the introduction of a "rainfall duration" (time span) parameter\n
    (-) the exposure of the flow depth parameter\n
    [[A more detailed description of the model, its modifications, and model application is provided by Setiawan (2012), chapter 6.[Currently, a number of additional grid datasets are outputted to facilitate model evaluation. This can be easily changed within the source code.[[[References:[[Morgan, R.P.C. (2001)]: A simple approach to soil loss prediction: a revised Morgan-Morgan-Finney model. Catena 44: 305-322.[[[Morgan, R.P.C., Duzant, J.H. (2008)]: Modified MMF (Morgan-Morgan-Finney) model for evaluating effects of crops and vegetation cover on soil erosion. Earth Surf. Process. Landforms 32: 90-106.[[[[Setiawan, M. A. (2012)]: Integrated Soil Erosion Management in the upper Serayu Watershed, Wonosobo District, Central Java Province, Indonesia. Dissertation at the Faculty of Geo- and Atmospheric Sciences of the University of Innsbruck, Austria.](http://sourceforge.net/projects/saga-gis/files/SAGA%20-%20Documentation/Modules/MMF-SAGA_Setiawan.pdf)[[\n
    Arguments
    ----------
    - DTM [`input grid`] : Digital Terrain Model. DTM, digital terrain model [m]
    - S [`input grid`] : Slope. S, slope [rad]
    - PI [`input grid`] : Permanent Interception. PI, permanent interception expressed as the proportion [between 0-1] of rainfall
    - CC [`input grid`] : Canopy Cover. CC, canopy cover expressed as a portion [between 0-1] of the soil surface protected by vegetation or crop
    - PH [`input grid`] : Plant Height. PH, plant height [m], representing the effective height from which raindrops fall from the crop or vegetation
    - ETEO [`input grid`] : Ratio Evapotranspiration. Et/Eo, ratio of actual to potential evapotranspiration
    - GC [`input grid`] : Ground cover. GC, Ground cover expressed as a portion [between 0-1] of the soil surface protected by vegetation or crop cover on the ground
    - D [`input grid`] : Diameter plant elements. D, Average diameter [m] of the individual plants elements (stem, leaves) at the ground surface
    - NV [`input grid`] : Number plant elements. NV, Number of plant elements per unit area [number/unit area] at the ground surface
    - MS [`input grid`] : Soil moisture (at FC). MS, Soil moisture at field capacity [% w/w]
    - BD [`input grid`] : Bulk density top layer. BD, Bulk density of the top layer [Mg/m3]
    - EHD [`input grid`] : Effective hydrological depth. EHD, Effective hydrological depth of the soil [m]
    - LP [`input grid`] : Sat. lateral permeability. LP, Saturated lateral permeability of the soil [m/day]
    - PER_C [`input grid`] : Percentage clays. c, Percentage clays [%]
    - PER_Z [`input grid`] : Percentage silt. z, Percentage silt [%]
    - PER_S [`input grid`] : Percentage sand. s, Percentage sand [%]
    - ST [`input grid`] : Percentage rock fragments. ST, Percentage rock fragments on the soil surface [%]
    - CHANNEL [`optional input grid`] : Channel Network. Channel network, all other cells NoData
    - RFR [`optional input grid`] : Surface roughness. RFR, Surface roughness [cm/m]. In case the surface roughness is not provided as input, v_flow_t is set to 1.0, i.e. natural soil surface roughness is not accounted for.
    - TAB_METEO [`optional input table`] : Meteorological data. Meteorological data for multiple timestep modelling [model step (day); temperature (Celsius); rainfall (mm), rainfall intensity (mm/h); rainfall duration (day); timespan (days)]
    - IF [`output grid`] : Interflow. IF
    - Q [`output grid`] : Mean runoff. Q, estimation of mean runoff [mm]
    - SL [`output grid`] : Mean soil loss. SL, estimation of mean soil loss [kg]
    - Rf [`output grid`] : Effective Rainfall. Rf
    - KE [`output grid`] : Total Kinetic Energy. KE
    - Rc [`output grid`] : Soil moisture storage capacity. Rc
    - TCC [`output grid`] : Transport Capacity Clay. TCc
    - TCZ [`output grid`] : Transport Capacity Silt. TCz
    - TCS [`output grid`] : Transport Capacity Sand. SLs
    - Gc [`output grid`] : Available Clay. Gc
    - Gz [`output grid`] : Available Silt. Gz
    - Gs [`output grid`] : Available Sand. Gs
    - SLC [`output grid`] : Sediment Balance Clay. SLc
    - SLZ [`output grid`] : Sediment Balance Silt. SLz
    - SLS [`output grid`] : Sediment Balance Sand. SLs
    - TCONDC [`output grid`] : Transport Condition Clay. Sediment Limited [0], Transport Limited (SL = TC) [1], Transport Limited (SL = G) [2]
    - TCONDZ [`output grid`] : Transport Condition Silt. Sediment Limited [0], Transport Limited (SL = TC) [1], Transport Limited (SL = G) [2]
    - TCONDS [`output grid`] : Transport Condition Sand. Sediment Limited [0], Transport Limited (SL = TC) [1], Transport Limited (SL = G) [2]
    - W_UP [`output grid`] : Upslope Flow Width. W_up
    - OUT_PATH [`file path`] : Output file path. Full path to the directory for the output grids of each model step
    - INTERFLOW [`boolean`] : Simulate Interflow. Default: 1 Simulate interflow
    - T [`floating point number`] : Mean temperature. Default: 18.000000 T, mean temperature [degree C]
    - TIMESPAN [`integer number`] : Timespan (days). Minimum: 1 Maximum: 365 Default: 30 The number of days to model.
    - R [`floating point number`] : Rainfall. Default: 200.000000 R, height of precipitation in timespan [mm]
    - I [`floating point number`] : Rainfall intensity. Default: 20.000000 I, rainfall intensity [mm/h]
    - Rn [`floating point number`] : Rainfall Duration. Default: 20.000000 Rn, number of rain days in timespan [-]
    - KE_I_METHOD [`choice`] : Relationship KE - I. Available Choices: [0] North America east of Rocky Mountains (Wischmeier & Smith 1978) [1] North-western Europe (Marshall & Palmer) [2] Mediterranean-type climates (Zanchi & Torri 1980) [3] Western Mediterranean (Coutinho & Tomas 1995) [4] Tropical climates (Hudson 1965) [5] Eastern Asia (Onaga et al. 1998) [6] Southern hemisphere climates (Rosewell 1986) [7] Bogor, West-Java, Indonesia (McISaac 1990) Default: 0 Relationship between kinetic energy (KE) and rainfall intensity (I)
    - FLOWD_VA [`floating point number`] : Flow Depth (actual flow velocity). Minimum: 0.000000 Default: 0.005000 The flow depth used to calculate the actual flow velocity [m] (e.g. 0.005 unchannelled flow, 0.01 shallow rills, 0.25 deeper rills.
    - CHANNELTRANSPORT [`boolean`] : Route Soil along Channel Network. Default: 0 Route soil loss along channel network to outlet

    - Verbose [`integer number`] : Verbosity level, 0=silent, 1=tool name and success notification, 2=complete tool output.\n
    Returns
    ----------
    `boolean` : `True` on success, `False` on failure.
    '''
    Tool = Tool_Wrapper('sim_erosion', '0', 'MMF-SAGA Soil Erosion Model')
    if Tool.is_Okay():
        Tool.Set_Input ('DTM', DTM)
        Tool.Set_Input ('S', S)
        Tool.Set_Input ('PI', PI)
        Tool.Set_Input ('CC', CC)
        Tool.Set_Input ('PH', PH)
        Tool.Set_Input ('EtEo', ETEO)
        Tool.Set_Input ('GC', GC)
        Tool.Set_Input ('D', D)
        Tool.Set_Input ('NV', NV)
        Tool.Set_Input ('MS', MS)
        Tool.Set_Input ('BD', BD)
        Tool.Set_Input ('EHD', EHD)
        Tool.Set_Input ('LP', LP)
        Tool.Set_Input ('PER_C', PER_C)
        Tool.Set_Input ('PER_Z', PER_Z)
        Tool.Set_Input ('PER_S', PER_S)
        Tool.Set_Input ('ST', ST)
        Tool.Set_Input ('CHANNEL', CHANNEL)
        Tool.Set_Input ('RFR', RFR)
        Tool.Set_Input ('TAB_METEO', TAB_METEO)
        Tool.Set_Output('IF', IF)
        Tool.Set_Output('Q', Q)
        Tool.Set_Output('SL', SL)
        Tool.Set_Output('Rf', Rf)
        Tool.Set_Output('KE', KE)
        Tool.Set_Output('Rc', Rc)
        Tool.Set_Output('TCc', TCC)
        Tool.Set_Output('TCz', TCZ)
        Tool.Set_Output('TCs', TCS)
        Tool.Set_Output('Gc', Gc)
        Tool.Set_Output('Gz', Gz)
        Tool.Set_Output('Gs', Gs)
        Tool.Set_Output('SLc', SLC)
        Tool.Set_Output('SLz', SLZ)
        Tool.Set_Output('SLs', SLS)
        Tool.Set_Output('TCONDc', TCONDC)
        Tool.Set_Output('TCONDz', TCONDZ)
        Tool.Set_Output('TCONDs', TCONDS)
        Tool.Set_Output('W_up', W_UP)
        Tool.Set_Option('OUT_PATH', OUT_PATH)
        Tool.Set_Option('INTERFLOW', INTERFLOW)
        Tool.Set_Option('T', T)
        Tool.Set_Option('TIMESPAN', TIMESPAN)
        Tool.Set_Option('R', R)
        Tool.Set_Option('I', I)
        Tool.Set_Option('Rn', Rn)
        Tool.Set_Option('KE_I_METHOD', KE_I_METHOD)
        Tool.Set_Option('FLOWD_VA', FLOWD_VA)
        Tool.Set_Option('CHANNELTRANSPORT', CHANNELTRANSPORT)
        return Tool.Execute(Verbose)
    return False

def run_tool_sim_erosion_0(
    DTM=None,
    S=None,
    PI=None,
    CC=None,
    PH=None,
    ETEO=None,
    GC=None,
    D=None,
    NV=None,
    MS=None,
    BD=None,
    EHD=None,
    LP=None,
    PER_C=None,
    PER_Z=None,
    PER_S=None,
    ST=None,
    CHANNEL=None,
    RFR=None,
    TAB_METEO=None,
    IF=None,
    Q=None,
    SL=None,
    Rf=None,
    KE=None,
    Rc=None,
    TCC=None,
    TCZ=None,
    TCS=None,
    Gc=None,
    Gz=None,
    Gs=None,
    SLC=None,
    SLZ=None,
    SLS=None,
    TCONDC=None,
    TCONDZ=None,
    TCONDS=None,
    W_UP=None,
    OUT_PATH=None,
    INTERFLOW=None,
    T=None,
    TIMESPAN=None,
    R=None,
    I=None,
    Rn=None,
    KE_I_METHOD=None,
    FLOWD_VA=None,
    CHANNELTRANSPORT=None, Verbose=2):
    '''
    MMF-SAGA Soil Erosion Model
    ----------
    [sim_erosion.0]\n
    Soil erosion modelling with a modified MMF (Morgan-Morgan-Finney) model (Morgan & Duzant 2008).[This tool is called MMF-SAGA because some things have been implemented differently compared to the original publication. The most important are:[[\n
    (-) the extension of the model to support spatially distributed modelling\n
    (-) the introduction of a "Channel Network" layer\n
    (-) the introduction of a "rainfall duration" (time span) parameter\n
    (-) the exposure of the flow depth parameter\n
    [[A more detailed description of the model, its modifications, and model application is provided by Setiawan (2012), chapter 6.[Currently, a number of additional grid datasets are outputted to facilitate model evaluation. This can be easily changed within the source code.[[[References:[[Morgan, R.P.C. (2001)]: A simple approach to soil loss prediction: a revised Morgan-Morgan-Finney model. Catena 44: 305-322.[[[Morgan, R.P.C., Duzant, J.H. (2008)]: Modified MMF (Morgan-Morgan-Finney) model for evaluating effects of crops and vegetation cover on soil erosion. Earth Surf. Process. Landforms 32: 90-106.[[[[Setiawan, M. A. (2012)]: Integrated Soil Erosion Management in the upper Serayu Watershed, Wonosobo District, Central Java Province, Indonesia. Dissertation at the Faculty of Geo- and Atmospheric Sciences of the University of Innsbruck, Austria.](http://sourceforge.net/projects/saga-gis/files/SAGA%20-%20Documentation/Modules/MMF-SAGA_Setiawan.pdf)[[\n
    Arguments
    ----------
    - DTM [`input grid`] : Digital Terrain Model. DTM, digital terrain model [m]
    - S [`input grid`] : Slope. S, slope [rad]
    - PI [`input grid`] : Permanent Interception. PI, permanent interception expressed as the proportion [between 0-1] of rainfall
    - CC [`input grid`] : Canopy Cover. CC, canopy cover expressed as a portion [between 0-1] of the soil surface protected by vegetation or crop
    - PH [`input grid`] : Plant Height. PH, plant height [m], representing the effective height from which raindrops fall from the crop or vegetation
    - ETEO [`input grid`] : Ratio Evapotranspiration. Et/Eo, ratio of actual to potential evapotranspiration
    - GC [`input grid`] : Ground cover. GC, Ground cover expressed as a portion [between 0-1] of the soil surface protected by vegetation or crop cover on the ground
    - D [`input grid`] : Diameter plant elements. D, Average diameter [m] of the individual plants elements (stem, leaves) at the ground surface
    - NV [`input grid`] : Number plant elements. NV, Number of plant elements per unit area [number/unit area] at the ground surface
    - MS [`input grid`] : Soil moisture (at FC). MS, Soil moisture at field capacity [% w/w]
    - BD [`input grid`] : Bulk density top layer. BD, Bulk density of the top layer [Mg/m3]
    - EHD [`input grid`] : Effective hydrological depth. EHD, Effective hydrological depth of the soil [m]
    - LP [`input grid`] : Sat. lateral permeability. LP, Saturated lateral permeability of the soil [m/day]
    - PER_C [`input grid`] : Percentage clays. c, Percentage clays [%]
    - PER_Z [`input grid`] : Percentage silt. z, Percentage silt [%]
    - PER_S [`input grid`] : Percentage sand. s, Percentage sand [%]
    - ST [`input grid`] : Percentage rock fragments. ST, Percentage rock fragments on the soil surface [%]
    - CHANNEL [`optional input grid`] : Channel Network. Channel network, all other cells NoData
    - RFR [`optional input grid`] : Surface roughness. RFR, Surface roughness [cm/m]. In case the surface roughness is not provided as input, v_flow_t is set to 1.0, i.e. natural soil surface roughness is not accounted for.
    - TAB_METEO [`optional input table`] : Meteorological data. Meteorological data for multiple timestep modelling [model step (day); temperature (Celsius); rainfall (mm), rainfall intensity (mm/h); rainfall duration (day); timespan (days)]
    - IF [`output grid`] : Interflow. IF
    - Q [`output grid`] : Mean runoff. Q, estimation of mean runoff [mm]
    - SL [`output grid`] : Mean soil loss. SL, estimation of mean soil loss [kg]
    - Rf [`output grid`] : Effective Rainfall. Rf
    - KE [`output grid`] : Total Kinetic Energy. KE
    - Rc [`output grid`] : Soil moisture storage capacity. Rc
    - TCC [`output grid`] : Transport Capacity Clay. TCc
    - TCZ [`output grid`] : Transport Capacity Silt. TCz
    - TCS [`output grid`] : Transport Capacity Sand. SLs
    - Gc [`output grid`] : Available Clay. Gc
    - Gz [`output grid`] : Available Silt. Gz
    - Gs [`output grid`] : Available Sand. Gs
    - SLC [`output grid`] : Sediment Balance Clay. SLc
    - SLZ [`output grid`] : Sediment Balance Silt. SLz
    - SLS [`output grid`] : Sediment Balance Sand. SLs
    - TCONDC [`output grid`] : Transport Condition Clay. Sediment Limited [0], Transport Limited (SL = TC) [1], Transport Limited (SL = G) [2]
    - TCONDZ [`output grid`] : Transport Condition Silt. Sediment Limited [0], Transport Limited (SL = TC) [1], Transport Limited (SL = G) [2]
    - TCONDS [`output grid`] : Transport Condition Sand. Sediment Limited [0], Transport Limited (SL = TC) [1], Transport Limited (SL = G) [2]
    - W_UP [`output grid`] : Upslope Flow Width. W_up
    - OUT_PATH [`file path`] : Output file path. Full path to the directory for the output grids of each model step
    - INTERFLOW [`boolean`] : Simulate Interflow. Default: 1 Simulate interflow
    - T [`floating point number`] : Mean temperature. Default: 18.000000 T, mean temperature [degree C]
    - TIMESPAN [`integer number`] : Timespan (days). Minimum: 1 Maximum: 365 Default: 30 The number of days to model.
    - R [`floating point number`] : Rainfall. Default: 200.000000 R, height of precipitation in timespan [mm]
    - I [`floating point number`] : Rainfall intensity. Default: 20.000000 I, rainfall intensity [mm/h]
    - Rn [`floating point number`] : Rainfall Duration. Default: 20.000000 Rn, number of rain days in timespan [-]
    - KE_I_METHOD [`choice`] : Relationship KE - I. Available Choices: [0] North America east of Rocky Mountains (Wischmeier & Smith 1978) [1] North-western Europe (Marshall & Palmer) [2] Mediterranean-type climates (Zanchi & Torri 1980) [3] Western Mediterranean (Coutinho & Tomas 1995) [4] Tropical climates (Hudson 1965) [5] Eastern Asia (Onaga et al. 1998) [6] Southern hemisphere climates (Rosewell 1986) [7] Bogor, West-Java, Indonesia (McISaac 1990) Default: 0 Relationship between kinetic energy (KE) and rainfall intensity (I)
    - FLOWD_VA [`floating point number`] : Flow Depth (actual flow velocity). Minimum: 0.000000 Default: 0.005000 The flow depth used to calculate the actual flow velocity [m] (e.g. 0.005 unchannelled flow, 0.01 shallow rills, 0.25 deeper rills.
    - CHANNELTRANSPORT [`boolean`] : Route Soil along Channel Network. Default: 0 Route soil loss along channel network to outlet

    - Verbose [`integer number`] : Verbosity level, 0=silent, 1=tool name and success notification, 2=complete tool output.\n
    Returns
    ----------
    `boolean` : `True` on success, `False` on failure.
    '''
    Tool = Tool_Wrapper('sim_erosion', '0', 'MMF-SAGA Soil Erosion Model')
    if Tool.is_Okay():
        Tool.Set_Input ('DTM', DTM)
        Tool.Set_Input ('S', S)
        Tool.Set_Input ('PI', PI)
        Tool.Set_Input ('CC', CC)
        Tool.Set_Input ('PH', PH)
        Tool.Set_Input ('EtEo', ETEO)
        Tool.Set_Input ('GC', GC)
        Tool.Set_Input ('D', D)
        Tool.Set_Input ('NV', NV)
        Tool.Set_Input ('MS', MS)
        Tool.Set_Input ('BD', BD)
        Tool.Set_Input ('EHD', EHD)
        Tool.Set_Input ('LP', LP)
        Tool.Set_Input ('PER_C', PER_C)
        Tool.Set_Input ('PER_Z', PER_Z)
        Tool.Set_Input ('PER_S', PER_S)
        Tool.Set_Input ('ST', ST)
        Tool.Set_Input ('CHANNEL', CHANNEL)
        Tool.Set_Input ('RFR', RFR)
        Tool.Set_Input ('TAB_METEO', TAB_METEO)
        Tool.Set_Output('IF', IF)
        Tool.Set_Output('Q', Q)
        Tool.Set_Output('SL', SL)
        Tool.Set_Output('Rf', Rf)
        Tool.Set_Output('KE', KE)
        Tool.Set_Output('Rc', Rc)
        Tool.Set_Output('TCc', TCC)
        Tool.Set_Output('TCz', TCZ)
        Tool.Set_Output('TCs', TCS)
        Tool.Set_Output('Gc', Gc)
        Tool.Set_Output('Gz', Gz)
        Tool.Set_Output('Gs', Gs)
        Tool.Set_Output('SLc', SLC)
        Tool.Set_Output('SLz', SLZ)
        Tool.Set_Output('SLs', SLS)
        Tool.Set_Output('TCONDc', TCONDC)
        Tool.Set_Output('TCONDz', TCONDZ)
        Tool.Set_Output('TCONDs', TCONDS)
        Tool.Set_Output('W_up', W_UP)
        Tool.Set_Option('OUT_PATH', OUT_PATH)
        Tool.Set_Option('INTERFLOW', INTERFLOW)
        Tool.Set_Option('T', T)
        Tool.Set_Option('TIMESPAN', TIMESPAN)
        Tool.Set_Option('R', R)
        Tool.Set_Option('I', I)
        Tool.Set_Option('Rn', Rn)
        Tool.Set_Option('KE_I_METHOD', KE_I_METHOD)
        Tool.Set_Option('FLOWD_VA', FLOWD_VA)
        Tool.Set_Option('CHANNELTRANSPORT', CHANNELTRANSPORT)
        return Tool.Execute(Verbose)
    return False

